from django import VERSION as DJANGO_VERSION
from django.conf import settings
from django.contrib.postgres.fields import (
    ArrayField, HStoreField, IntegerRangeField,
    DateRangeField, DateTimeRangeField)
from django.contrib.postgres.operations import (
    HStoreExtension, UnaccentExtension)
from django.db import models, migrations


def extra_regular_available_fields():
    fields = []
    try:
        from django.db.models import JSONField
        fields.append(('json', JSONField(null=True, blank=True)))
    except ImportError:
        pass

    return fields


def extra_postgres_available_fields():
    fields = []
    try:
        # TODO Remove when Dj31 support is dropped
        from django.contrib.postgres.fields import FloatRangeField
        fields.append(('float_range', FloatRangeField(null=True, blank=True)))
    except ImportError:
        pass

    try:
        # TODO Add to module import when Dj31 is dropped
        from django.contrib.postgres.fields import DecimalRangeField
        fields.append(('decimal_range', DecimalRangeField(null=True, blank=True)))
    except ImportError:
        pass

    # Future proofing with Django 40 deprecation
    if DJANGO_VERSION[0] < 4:
        # TODO Remove when Dj40 support is dropped
        from django.contrib.postgres.fields import JSONField
        fields.append(('json', JSONField(null=True, blank=True)))

    return fields


class Migration(migrations.Migration):

    dependencies = [
        ('auth', '0001_initial'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Test',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('name', models.CharField(max_length=20)),
                ('public', models.BooleanField(default=False)),
                ('date', models.DateField(null=True, blank=True)),
                ('datetime', models.DateTimeField(null=True, blank=True)),
                ('owner', models.ForeignKey(blank=True, to=settings.AUTH_USER_MODEL, null=True, on_delete=models.SET_NULL)),
                ('permission', models.ForeignKey(blank=True, to='auth.Permission', null=True, on_delete=models.PROTECT)),
                ('a_float', models.FloatField(null=True, blank=True)),
                ('a_decimal', models.DecimalField(null=True, blank=True, max_digits=5, decimal_places=2)),
                ('bin', models.BinaryField(null=True, blank=True)),
                ('ip', models.GenericIPAddressField(null=True, blank=True)),
                ('duration', models.DurationField(null=True, blank=True)),
                ('uuid', models.UUIDField(null=True, blank=True)),
            ] + extra_regular_available_fields(),
            options={
                'ordering': ('name',),
            },
        ),
        migrations.CreateModel(
            name='TestParent',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('name', models.CharField(max_length=20)),
            ],
        ),
        migrations.CreateModel(
            name='TestChild',
            fields=[
                ('testparent_ptr', models.OneToOneField(parent_link=True, auto_created=True, primary_key=True, serialize=False, to='cachalot.TestParent', on_delete=models.CASCADE)),
                ('public', models.BooleanField(default=False)),
                ('permissions', models.ManyToManyField('auth.Permission', blank=True))
            ],
            bases=('cachalot.testparent',),
        ),
        HStoreExtension(),
        UnaccentExtension(),
        migrations.CreateModel(
            name='PostgresModel',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False,
                                        auto_created=True, primary_key=True)),
                ('int_array', ArrayField(
                    models.IntegerField(null=True, blank=True), size=3,
                    null=True, blank=True)),
                ('hstore', HStoreField(null=True, blank=True)),
                ('int_range', IntegerRangeField(null=True, blank=True)),
                ('date_range', DateRangeField(null=True, blank=True)),
                ('datetime_range', DateTimeRangeField(null=True, blank=True)),
            ] + extra_postgres_available_fields(),
        ),
        migrations.RunSQL('CREATE TABLE cachalot_unmanagedmodel '
                          '(id SERIAL PRIMARY KEY, name VARCHAR(50));'),
    ]



from unittest import skipIf

from django.conf import settings
from django.contrib.auth.models import User
from django.db import DEFAULT_DB_ALIAS, transaction
from django.test import TransactionTestCase

from ..api import invalidate
from ..signals import post_invalidation

from .models import Test


class SignalsTestCase(TransactionTestCase):
    databases = set(settings.DATABASES.keys())

    def test_table_invalidated(self):
        l = []

        def receiver(sender, **kwargs):
            db_alias = kwargs['db_alias']
            l.append((sender, db_alias))

        post_invalidation.connect(receiver)
        self.assertListEqual(l, [])
        list(Test.objects.all())
        self.assertListEqual(l, [])
        Test.objects.create(name='test1')
        self.assertListEqual(l, [('cachalot_test', DEFAULT_DB_ALIAS)])
        post_invalidation.disconnect(receiver)

        del l[:]  # Empties the list
        post_invalidation.connect(receiver, sender=User._meta.db_table)
        Test.objects.create(name='test2')
        self.assertListEqual(l, [])
        User.objects.create_user('user')
        self.assertListEqual(l, [('auth_user', DEFAULT_DB_ALIAS)])
        post_invalidation.disconnect(receiver, sender=User._meta.db_table)

    def test_table_invalidated_in_transaction(self):
        """
        Checks that the ``post_invalidation`` signal is triggered only after
        the end of a transaction.
        """
        l = []

        def receiver(sender, **kwargs):
            db_alias = kwargs['db_alias']
            l.append((sender, db_alias))

        post_invalidation.connect(receiver)

        self.assertListEqual(l, [])
        with transaction.atomic():
            Test.objects.create(name='test1')
            self.assertListEqual(l, [])
        self.assertListEqual(l, [('cachalot_test', DEFAULT_DB_ALIAS)])

        del l[:]  # Empties the list
        self.assertListEqual(l, [])
        with transaction.atomic():
            Test.objects.create(name='test2')
            with transaction.atomic():
                Test.objects.create(name='test3')
                self.assertListEqual(l, [])
            self.assertListEqual(l, [])
        self.assertListEqual(l, [('cachalot_test', DEFAULT_DB_ALIAS)])

        post_invalidation.disconnect(receiver)

    def test_table_invalidated_once_per_transaction_or_invalidate(self):
        """
        Checks that the ``post_invalidation`` signal is triggered only after
        the end of a transaction.
        """
        l = []

        def receiver(sender, **kwargs):
            db_alias = kwargs['db_alias']
            l.append((sender, db_alias))

        post_invalidation.connect(receiver)

        self.assertListEqual(l, [])
        with transaction.atomic():
            Test.objects.create(name='test1')
            self.assertListEqual(l, [])
            Test.objects.create(name='test2')
            self.assertListEqual(l, [])
        self.assertListEqual(l, [('cachalot_test', DEFAULT_DB_ALIAS)])

        del l[:]  # Empties the list
        self.assertListEqual(l, [])
        invalidate(Test, db_alias=DEFAULT_DB_ALIAS)
        self.assertListEqual(l, [('cachalot_test', DEFAULT_DB_ALIAS)])

        del l[:]  # Empties the list
        self.assertListEqual(l, [])
        with transaction.atomic():
            invalidate(Test, db_alias=DEFAULT_DB_ALIAS)
            self.assertListEqual(l, [])
        self.assertListEqual(l, [('cachalot_test', DEFAULT_DB_ALIAS)])

        post_invalidation.disconnect(receiver)

    @skipIf(len(settings.DATABASES) == 1,
            'We can’t change the DB used since there’s only one configured')
    def test_table_invalidated_multi_db(self):
        db_alias2 = next(alias for alias in settings.DATABASES
                         if alias != DEFAULT_DB_ALIAS)
        l = []

        def receiver(sender, **kwargs):
            db_alias = kwargs['db_alias']
            l.append((sender, db_alias))

        post_invalidation.connect(receiver)
        self.assertListEqual(l, [])
        Test.objects.using(DEFAULT_DB_ALIAS).create(name='test')
        self.assertListEqual(l, [('cachalot_test', DEFAULT_DB_ALIAS)])
        Test.objects.using(db_alias2).create(name='test')
        self.assertListEqual(l, [
            ('cachalot_test', DEFAULT_DB_ALIAS),
            ('cachalot_test', db_alias2)])
        post_invalidation.disconnect(receiver)
